
#include <inttypes.h>
#include "network_config.hh"

// TODO: replace the remaining c string functions by something more advanced...

namespace network_config
{
  typedef std::string::size_type str_index;

  std::string Number2String (int number, const char* format="%d")
  {
    char buffer[64];
    sprintf (buffer, format, number);
    return std::string (buffer);
  };

  bool String2Number (const std::string& str, int* result)
  {
    long int res;
    const char* ptr = str.c_str ();
    char* endptr;
    res=strtol(ptr, &endptr, 0);
    *result=res;
    return (endptr != ptr && *endptr != 0);
  }


  void DeQuote (std::string& str)
  {
    unsigned int l = str.length ();
    if ((l > 1) && (str[0] == str[l-1])
	&&(str[0] == '"' || str[0] == '\'') )
      str = str.substr (1, l-2);
  }

  std::string Quote (std::string& str)
  { // todo: escape quotes
    return std::string ("\'") + str + std::string ("\'");
  }

  IP::IP ()
  {
    SetDefault ();
  }

  void IP::SetDefault ()
  {
    address[0]=127;
    address[1]=0;
    address[2]=0;
    address[3]=1;
  }

  bool IP::FromString (const std::string& str)
  {
    int addr [4];
    if (sscanf (str.c_str (), "%d.%d.%d.%d", addr, addr + 1, addr + 2, addr + 3) != 4) {
      SetDefault ();
      return false;
    }

    for (unsigned int i =0; i < 4; i++) {
      if (addr [i] < 0 || addr [i] > 255) {
	SetDefault ();
	return false;
      }
      address[i] = (block) addr[i]; 
    }

    return true;
  }

  std::string IP::ToString ()
  {
    char buffer [64];
    sprintf (buffer, "%d.%d.%d.%d", (int) address[0], (int) address[1], (int) address[2], (int) address[3]);
    return std::string ((char *) buffer);
  }

  CIDR::CIDR ()
  {
    netmask_bits = 8;
  }

  std::string CIDR::ToString ()
  {
    char buffer [64];
    sprintf (buffer, "%d.%d.%d.%d/%d", (int) address[0], (int) address[1], (int) address[2], (int) address[3], netmask_bits);
    return std::string ((char *) buffer);
  }

  bool CIDR::FromString (const std::string& str)
  {
    str_index delimiter = str.find_first_of("/");
    bool error = true;
    std::string ip_part;

    if (delimiter != std::string::npos) {
      ip_part=str.substr (0, delimiter);
      std::string delim_part=str.substr (delimiter);
      if (sscanf (delim_part.c_str (), "/%d", &netmask_bits) == 1) {
	if (netmask_bits >= 0 && netmask_bits <= 32)
	  error = false;
	else
	  netmask_bits=24;
      }
    } else
      ip_part=str;
    
    return (IP::FromString (ip_part) && !error);
  }

  
  template <typename CHILDT> CHILDT* getChild (std::vector <kvtree::kvNode*>& child_vec, kvtree::node_type type, unsigned int no)
  {
    unsigned int seen = 0;
    unsigned int no_childs = child_vec.size ();
    for (unsigned int i = 0; i < no_childs; i++)
      if (child_vec[i] -> Type () == type) {
	if (seen++ == no)
	  return (CHILDT*) child_vec[i];
      }
    return (CHILDT*) 0;
  }

  template <typename SUBT> SUBT* getSubsection (kvtree::node_type type, kvtree::Group* parent, bool autocreate) {
    SUBT* ret = getChild<SUBT> (parent -> childs, type, 0);
    if ((!ret) && autocreate) {
      ret = new SUBT (parent);
      parent -> childs.push_back (ret);
    }
    return ret;
  }


  InterfaceSection::InterfaceSection ( const std::string& iface, const std::string& prof, kvtree::Group* root) : kvtree::Group (root)
  {
    interface = iface;
    profile = prof;
    enabled = true;
  }

  AddressSubsection* InterfaceSection::GetAddressSubsection (bool autocreate)
  {
    return getSubsection <AddressSubsection> (kvtree::node_types::address_subsection, this, autocreate);
  }

  WirelessSubsection* InterfaceSection::GetWirelessSubsection (bool autocreate)
  {
    return getSubsection <WirelessSubsection> (kvtree::node_types::wireless_subsection, this, autocreate);
  }

  PPPSubsection* InterfaceSection::GetPPPSubsection (bool autocreate)
  {
    return getSubsection <PPPSubsection> (kvtree::node_types::ppp_subsection, this, autocreate);
  }


  void InterfaceSection::Dump (std::ostream& stream)
  {
    if (!enabled)
      return;
    stream << "interface " << interface << "(" << profile << ")" << std::endl;
    DumpChilds (stream);
  }

  void IPLine::Set ()
  {
    line="ip ";
    line += address.ToString (); 
  }

  IP* NetmaskLine::GetIP ()
  {
    int nmb = ipl -> address.netmask_bits;
    uint32_t mask = 0;
    for (int i = 0; i < 32; i++)
      {
	mask *= 2;
	if (nmb) {
	  nmb--;
	  mask++;
	}
      }

    //    std::cout << (void*) mask << std::endl;

    for (int i = 3; i >= 0; i--)
      {
	faked.address[i] = (IP::block) (mask % 0x100);
	mask /= 0x100;
      }

    return &faked;
  }

  void NetmaskLine::Set ()
  {
    uint32_t mask = 0;
    for (int i = 0; i < 4; i++)
      {
	mask *= 0x100;
	mask += faked.address[i];
      }

    int nmb = 32;
    for (int i = 0; i < 32 && !(mask & (1 << i)); i++) {
      nmb--;
    }

    ipl -> address.netmask_bits = nmb;
    ipl -> Set ();
    line = ipl -> line;
  }


  void GatewayLine::Set ()
  {
    line="gw ";
    line += nexthop.ToString (); 
  }

  void SearchLine::Set ()
  {
    line="search ";
    line += domain; 
  }

  void NameserverLine::Set ()
  {
    line="nameserver ";
    line += nameserver.ToString (); 
  }


 
  AddressSubsection::AddressSubsection ( kvtree::Group* interface) : kvtree::Group (interface)
  {
    dhcp = false;
    skip_nameservers = false;
    skip_gateways = false;
  }

  IPLine* AddressSubsection::Address (unsigned int number)
  {
    return getChild<IPLine> (childs, kvtree::node_types::ip_setup, number);
  }

  GatewayLine* AddressSubsection::Gateway (unsigned int number)
  {
    return getChild<GatewayLine> (childs, kvtree::node_types::gw_setup, number);
  }

  SearchLine* AddressSubsection::Search (unsigned int number)
  {
    return getChild<SearchLine> (childs, kvtree::node_types::search, number);
  }

  NameserverLine* AddressSubsection::Nameserver (unsigned int number)
  {
    return getChild<NameserverLine> (childs, kvtree::node_types::nameserver, number);
  }

  void AddressSubsection::Dump (std::ostream& stream)
  {
    if (dhcp) {
      stream << "\tdhcp" << std::endl;
    } else {
      unsigned int no_childs = Subnodes ();
      for (unsigned int i=0; i < no_childs; i++) {
	kvtree::node_type type = childs[i] -> Type ();

	bool dump = (type != kvtree::node_types::nameserver || !skip_nameservers)
	  && (type != kvtree::node_types::gw_setup || !skip_gateways);

	if (dump)
	  childs[i] -> Dump (stream);
      }
    }
  }


  void ModeLine::Set ()
  {
    switch (mode) {
    case adhoc:
      line = "mode Ad-Hoc";
      break;
    case managed:
    default:
      line = "mode Managed";
    }
  }


  bool ModeLine::FromString (const std::string& str)
  {
    if (str == "Managed" or str == "managed")
      mode = managed;
    else if (str == "Ad-Hoc" || str == "ad-hoc")
      mode = adhoc;
    else
      return false;
    return true;
  }



  void ChannelLine::Set ()
  {
    line = "channel ";
    line += Number2String (channel);
  }
  
  bool ChannelLine::FromString (const std::string& str)
  {
    int ichannel;
    if (String2Number(str, &ichannel) || ichannel < 0) {
      channel = 1;
      return false;
    }

    channel = (unsigned int) ichannel;
    return true;
  }

  void EssidLine::Set ()
  {
    
    line = "essid ";
    if (essid != "") {
      line += Quote (essid);
    } else
      line += "any";
  }

  void KeyLine::Set ()
  {
    line = "key ";
    if (key != "") {
      line += Quote (key);
    } else
      line += "off";
  }

  WirelessSubsection::WirelessSubsection ( kvtree::Group* interface) : kvtree::Group (interface)
  {
  }


  ModeLine* WirelessSubsection::Mode ()
  {
    ModeLine* ret = getChild<ModeLine> (childs, kvtree::node_types::wireless_mode, 0);
    if (!ret) {
      ret = new ModeLine ("");
      ret -> Set ();
      childs.push_back (ret);
    }
    return ret;
  }

  ChannelLine* WirelessSubsection::Channel ()
  {
    ChannelLine* ret = getChild<ChannelLine> (childs, kvtree::node_types::wireless_channel, 0);
    if (!ret) {
      ret = new ChannelLine ("");
      ret -> Set ();
      childs.push_back (ret);
    }
    return ret;
  }

  EssidLine* WirelessSubsection::Essid ()
  {
    EssidLine* ret = getChild<EssidLine> (childs, kvtree::node_types::wireless_essid, 0);
    if (!ret) {
      ret = new EssidLine ("");
      ret -> Set ();
      childs.push_back (ret);
    }
    return ret;
  }

  KeyLine* WirelessSubsection::Key ()
  {
    KeyLine* ret = getChild<KeyLine> (childs, kvtree::node_types::wireless_key, 0);
    if (!ret) {
      ret = new KeyLine ("");
      ret -> Set ();
      childs.push_back (ret);
    }
    return ret;
  }


  void WirelessSubsection::Dump (std::ostream& stream)
  {
    bool skip_channel = (Mode() -> mode == ModeLine::managed);
    unsigned int no_childs = Subnodes ();
    for (unsigned int i=0; i < no_childs; i++) {
      kvtree::node_type type = childs[i] -> Type ();

      bool dump = (type != kvtree::node_types::wireless_channel || !skip_channel);
      if (dump)
	childs[i] -> Dump (stream);
    }   
  }



  void DeviceLine::Set ()
  {
    line = "ppp ";
    line += device;
    line += " debug";
    if (speed != 0)
      line += Number2String (speed);
  }

  bool DeviceLine::SpeedFromString (const std::string& speed_str)
  {
    int ispeed;
    if (String2Number(speed_str, &ispeed) || speed < 0) {
      speed = 0;
      return false;
    }

    speed = (unsigned int) ispeed;
    return true;
  }


  void ChatInitLine::Set ()
  {
    line = "chat_init ";
    line += Quote(init);
  }

  void ChatDialLine::Set ()
  {
    // TODO: quoting
    line = "chat_dial ";
    line += Quote(dial);
  }

  void OptionLine::Set ()
  {
    // TODO: quoting
    line = "ppp_option ";
    line += option;
    if (value != "")
      line += " " + Quote (value);
  }



  PPPSubsection::PPPSubsection ( kvtree::Group* interface) : kvtree::Group (interface)
  {
    pppoe = false;
    ppp_defaults = false;
    ppp_speed_defaults = false;
    ppp_chat_defaults = false;
  }


  OptionLine* PPPSubsection::Option (unsigned int number)
  {
    return getChild<OptionLine> (childs, kvtree::node_types::ppp_option, number);
  }

  OptionLine* PPPSubsection::Option (const std::string& name, bool autocreate)
  {
    unsigned int end = childs.size ();
    for (unsigned int i=0; i < end; i++)
      if (childs[i] -> Type () == kvtree::node_types::ppp_option) {
	OptionLine* cur = (OptionLine*) childs[i];
	if (cur -> option == name)
	  return cur;
      }

    if (autocreate) {
      OptionLine* created = new OptionLine ("");
      created -> option = name;
      created -> Set ();
      childs.push_back (created);
      return created;
    } else
      return 0;
  }


  DeviceLine* PPPSubsection::Device ()
  {
    DeviceLine* dev = getChild<DeviceLine> (childs, kvtree::node_types::ppp_device, 0);
    if (!dev) {
      dev = new DeviceLine ("");
      childs.push_back (dev);
    };
    return dev;
  }


  ChatInitLine* PPPSubsection::ChatInit ()
  {
    ChatInitLine* chl = getChild<ChatInitLine> (childs, kvtree::node_types::ppp_chat_init, 0);
    if (!chl) {
      chl = new ChatInitLine ("");
      childs.push_back (chl);
    };
    return chl;
  }

  ChatDialLine* PPPSubsection::ChatDial ()
  {
    ChatDialLine* chl = getChild<ChatDialLine> (childs, kvtree::node_types::ppp_chat_dial, 0);
    if (!chl) {
      chl = new ChatDialLine ("");
      childs.push_back (chl);
    };
    return chl;
  }


 
  void PPPSubsection::Dump (std::ostream& stream)
  {
    if (Device()) // paranoia
      Device () -> Dump (stream);
    if (pppoe)
      stream << "\tpppoe" << std::endl;
    if (ppp_defaults)
      stream << "\tppp_defaults" << std::endl;
    if (ppp_speed_defaults)
      stream << "\tppp_speed_defaults" << std::endl;
    if (ppp_chat_defaults)
      stream << "\tchat_defaults" << std::endl;

    unsigned int no_childs = Subnodes ();
    for (unsigned int i=0; i < no_childs; i++) {
      kvtree::node_type type = childs[i] -> Type ();
      bool dump = (type != kvtree::node_types::ppp_device);
      if (type == kvtree::node_types::ppp_option) {
	OptionLine* option = (OptionLine*) childs[i];
	if (option -> value == "") { // do not dump any empty option which expect a value
	  if (option -> option == "user" || option -> option == "password")
	    dump = false;
	}
      }

      if (pppoe && type == kvtree::node_types::ppp_chat_init)
	dump = false;
      if (pppoe && type == kvtree::node_types::ppp_chat_dial)
	dump = false;
	
      if (dump)
	childs[i] -> Dump (stream);
    }
  }
}
